#include "stdafx.h"
#include <windows.h>
#include <stdio.h>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>
#include <cstddef>
#include <ctime>
#include <conio.h>

extern HANDLE hSerial1;
extern char szBuff[];
extern DWORD dwBytesWrite;
extern DWORD dwBytesRead;

#define SAMP_SIZE 7
#define MON_RETRY_COUNT 4 // Was 10, seemed too long
#define MON_SLEEP 200

enum msg_type {SET_TEMP_MIN, SET_TEMP_MAX, SET_TEMP_NUM_SAMPLES, SET_MOTION_NUM_SAMPLES, 
	SET_PERIOD, GET_TEMP_SAMPLES, GET_MOTION_SAMPLES, GET_MIN_TEMP, GET_MAX_TEMP, 
	GET_PERIOD, GET_TEMP_NUM_SAMPLES, GET_MOTION_NUM_SAMPLES, SET_MOTION_THRESH, 
	GET_MOTION_THRESH, MON_ERROR, EXIT=-1, TIMEOUT=-5, ACK=-10};

bool get_input ( char *buffer, std::size_t size, int timeout )
{
  std::time_t start = std::time ( 0 );
  std::size_t n = 0;
  for ( ; ; ) {
    if ( n == 0 && std::difftime ( std::time ( 0 ), start ) >= timeout )
      return false;
    if ( _kbhit() ) {
      if ( n == size - 1 )
        break;
      char ch = (int)_getche();
      if ( ch == '\r' ) {
        buffer[n++] = '\n';
        break;
      }
      else
        buffer[n++] = ch;
    }
  }
  buffer[n] = '\0';
  return true;
} 

bool dev_send(char msg) {
    
	*szBuff = msg;
    if (!WriteFile(hSerial1, szBuff, sizeof(char), &dwBytesWrite, NULL)) {
        printf("write failed\n");
		return false;
    }
	else {
		return true;
	}
}

char dev_get(int &read_cnt) {
	int retries = MON_RETRY_COUNT;
	bool alert_processing = true;
	bool prev_was_a = false;

	read_cnt = 1; // Assum read, if timeout set to zero
	// Retry several times since device is slow.  Sleep a bit at the 
	// end.  Don't want to leave before device has sent characters
	// on slow interface so be patient.
	while ((retries > 0) || alert_processing) {

		if (retries == 0) {
			// Means we haven't gotten any data, so be done processing
			// and return a zero count (nothing read)
			read_cnt = 0;
			alert_processing = false;
			return szBuff[0];
		}
		
	    if (!ReadFile(hSerial1, szBuff, 1, &dwBytesRead, NULL)) {
            printf("read failed\n");
		    return 'e';
        }
		// We have some character from the device
	    else if (dwBytesRead > 0) {
			// Must consume all alerts from device in this module.
			// Device will only send a, m, l, or h as an alert.  Any
			// other character can be passed on.
			if (szBuff[0] == 'a' || szBuff[0] == 'm' || szBuff[0] == 'h' || szBuff[0] == 'l') {
				if (szBuff[0] == 'a') {
					prev_was_a = true;
					printf("\n\nALERT RECEIVED: ");
					dev_send('k');
				}
				else if (szBuff[0] == 'm') {
					prev_was_a = false;
					printf(" MOTION DETECTED\n\n");
					dev_send('k');
				}
				else if (szBuff[0] == 'h') {
					prev_was_a = false;
					printf(" TEMPERATURE ABOVE THRESHOLD\n\n");
					dev_send('k');
				}
				else if (szBuff[0] == 'l') {
					prev_was_a = false;
					printf(" TEMPERATURE BELOW THRESHOLD\n\n");
					dev_send('k');
				}	
			}
			// Nothing related to alerts now, so normal handling - return character
			else {
				// Catch case where last character received
				// as an 'a' and next one was not.
				if (prev_was_a) {
					// Handshaking error?
					printf(" UNKNOWN alert %c\n\n",szBuff[0]);
					dev_send('e');
				}
				// Return character to caller since non-error related
				return szBuff[0];
			}
	    }
		else {
			retries--; // Decrement count since no data
			// Sleep a bit and see if host has sent anything for retry
			Sleep(MON_SLEEP);
		}
	}
	return szBuff[0];
}

bool dev_connect() {
	int read_cnt;
	dev_send('c');
	if (dev_get(read_cnt) == 'k') {
		printf("Connection with device established\n");
		return true;
	}
	else {
		printf("Device did not respond to connection request\n");
		return false;
	}
}

// Need to clean this up I think.  It should only ever process
// alerts.  It would be an error to get any non-alert I believe
// since all commands should be handshaked.
bool listen_for_alerts(void) {
	int read_cnt;
	char msg_buf = dev_get(read_cnt);
	if (read_cnt > 0) {
		printf("Hm.  Stole someone's command??  Got %c\n", msg_buf);
		// FIXME - should I call parse_cmd()??
		return false;
	}
	else {
		return true;
	}
}

double get_sample(void) {
    char char_sample[SAMP_SIZE+1];
	int read_cnt;
	char_sample[SAMP_SIZE] = '\0';
    for (int i=0; i<SAMP_SIZE; ++i) {
        char_sample[i] = dev_get(read_cnt);
    }
	// FIXME - need error checking here on sample.  Try
	// printing it to a string and make sure it will fit in
	// 7 characters or don't allow it to be set.
    return atof(char_sample);
}

bool get_data(char msg, double &data) {
	char dev_msg;
	int read_cnt;
	// Consume any alerts that may have arrived
	// otherwise handshaking protocol is off
	while (!listen_for_alerts());
	dev_send(msg);
	// Device should ack
    dev_msg = dev_get(read_cnt);
	if (dev_msg == 'k') {
		data = get_sample();
		return true;
	}
	else {
		printf("Expected ACK got - %c\n",dev_msg);
		return false;
	}
}

bool get_array(char msg, std::vector<double> &data) {
	char dev_msg;
	int num_samples;
	int read_cnt;
	// Consume any alerts that may have arrived
	// otherwise handshaking protocol is off
	while (!listen_for_alerts());
	dev_send(msg);
	// Device should ack
    dev_msg = dev_get(read_cnt);
	if (dev_msg == 'k') {
		num_samples = (int)get_sample();
		for (int i=0; i < num_samples; ++i) {
		    data.push_back(get_sample());
		}
		return true;
	}
	else {
		return false;
	}
}

bool convert_sample(double sample, std::string &toString) {
    std::ostringstream ss;

    ss.precision(5);
    ss.width(7); ss << sample;
    toString = ss.str();
    // Can't send strings bigger than 7 characters.  If that happens
    // it was user error!
    if (toString.size() > 7) {
        return false;
    }
    return true;
}

bool send_data(double sample) {
	// Somewhat kludgy way of creating a fixed width field so
	// I don't have to worry about parsing.  All samples are always
	// 7 characters wide and will be padded with spaces to ensure it.

    std::string mystring; 
    if (convert_sample(sample, mystring)) {
        for (size_t i=0; i < mystring.size(); ++i) {
		    dev_send(mystring[i]);
	    }
	    return true;
	}
	else {
	    dev_send('e');
		return false;
	}
}

char get_user_cmd(void) {
	char entry[256];
	printf(" 1) Set Alert Min Temperature Threshold\n");
	printf(" 2) Set Alert Max Temperature Threshold\n");
	printf(" 3) Set Number of Temperature Samples Stored\n");
	printf(" 4) Set Number of Motion Samples Stored\n");
	printf(" 5) Set Termperature Rport Sample Period in Seconds\n");
	printf(" 6) Get All Temperature Samples\n");
	printf(" 7) Get All Motion Samples\n");
	printf(" 8) Get Min Temperature Threshold Setting\n");
	printf(" 9) Get Max Temperature Threshold Setting\n");
	printf("10) Get Temperature Sample Period in Seconds\n");
	printf("11) Get Size of Temperature Buffer\n");
	printf("12) Get Size of Motion Alert Buffer\n");
	printf("13) Set X, Y and Z Motion Thresholds\n");
	printf("14) Get X, Y, and Z Motion Threshold Settings\n");
	printf("15) Send ACK (in case of comm problems)\n");
	printf("16) Exit Program\n");
	printf("\nEnter Your Choice (1-15): \n");
	// Wait 5 seconds for input then need to check for alerts
	if (!get_input(entry, 256, 5)) {
		return -5;  // Indicate a timeout
	}
	else {
		// User wants to exit
		if (atoi(entry) == 16) {
			return -1;
		}
		// Special entry just for debug to send ack's to
		// help flush commands from device expecting to be acked
		else if (atoi(entry) == 15) {
			return -10;
		}
		// User actually entered something
	    return atoi(entry)-1;
	}
}

void check_response(char msg) {
	if (msg == 'k') {
		printf("Device received command\n");
	}
	else if (msg == 'e') {
		printf("Error received from device\n");
	}
	else {
		printf("Unkown response from device %c\n", msg);
	}
}

bool parse_cmd(int msg) {
	int read_cnt;
	char entry[256];
	char dev_msg;
	double val;
	std::vector<double> array_vals;
	switch(msg) {
	case SET_TEMP_MIN:
		printf("Set Min Temp: ");
		std::cin.getline(entry,256);
	    val = atof(entry);
		// Consume any alerts that may have arrived
		// otherwise handshaking protocol is off
		while (!listen_for_alerts());
		dev_send('z');
		dev_msg = dev_get(read_cnt);
		check_response(dev_msg);
		send_data(val);
		dev_msg = dev_get(read_cnt);
		check_response(dev_msg);
		break;
	case SET_TEMP_MAX:
		printf("Set Max Temp: ");
		std::cin.getline(entry,256);
	    val = atof(entry);
		// Consume any alerts that may have arrived
		// otherwise handshaking protocol is off
		while (!listen_for_alerts());
		dev_send('y');
		dev_msg = dev_get(read_cnt);
		check_response(dev_msg);
		send_data(val);
		dev_msg = dev_get(read_cnt);
		check_response(dev_msg);
		break;
	case SET_TEMP_NUM_SAMPLES:
		printf("Set Num Temp Samples Stored: ");
		std::cin.getline(entry,256);
	    val = atof(entry);
		// Consume any alerts that may have arrived
		// otherwise handshaking protocol is off
		while (!listen_for_alerts());
		dev_send('x');
		dev_msg = dev_get(read_cnt);
		check_response(dev_msg);
		send_data(val);
		dev_msg = dev_get(read_cnt);
		check_response(dev_msg);
		break;
	case SET_MOTION_NUM_SAMPLES:
		printf("Set Num Motion Samples Stored: ");
		std::cin.getline(entry,256);
	    val = atof(entry);
		// Consume any alerts that may have arrived
		// otherwise handshaking protocol is off
		while (!listen_for_alerts());
		dev_send('w');
		dev_msg = dev_get(read_cnt);
		check_response(dev_msg);
		send_data(val);
		dev_msg = dev_get(read_cnt);
		check_response(dev_msg);
		break;
	case SET_PERIOD:
		printf("Set Temperature Sample Period (secs): ");
		std::cin.getline(entry,256);
	    val = atof(entry);
		// Consume any alerts that may have arrived
		// otherwise handshaking protocol is off
		while (!listen_for_alerts());
		dev_send('v');
		dev_msg = dev_get(read_cnt);
		check_response(dev_msg);
		send_data(val);
		dev_msg = dev_get(read_cnt);
		check_response(dev_msg);
		break;
	case SET_MOTION_THRESH:
		// Consume any alerts that may have arrived
		// otherwise handshaking protocol is off
		while (!listen_for_alerts());
		dev_send('u');
		dev_msg = dev_get(read_cnt);
		check_response(dev_msg);
		printf("Set X Motion Threshold: ");
		std::cin.getline(entry,256);
	    val = atof(entry);
		send_data(val);
		printf("Set Y Motion Threshold: ");
		std::cin.getline(entry,256);
	    val = atof(entry);
		send_data(val);
		printf("Set Z Motion Threshold: ");
		std::cin.getline(entry,256);
	    val = atof(entry);
		send_data(val);
		dev_msg = dev_get(read_cnt);
		check_response(dev_msg);
		break;
	case GET_TEMP_SAMPLES:
		if (get_array('t', array_vals)) {
			printf("Temperature Samples: \n");
			for (size_t i=0; i<array_vals.size(); ++i) {
				printf("Sample %d = %f\n", i, array_vals[i]);
			}
		}
		else {
			printf("ERROR: Ack not received from device\n");
		}
		break;
	case GET_MOTION_SAMPLES:
		if (get_array('s', array_vals)) {
			printf("Motion Samples: \n");
			for (size_t i=0; i<array_vals.size(); ++i) {
				printf("Sample %d = %f\n", i, array_vals[i]);
			}
		}
		else {
			printf("ERROR: Ack not received from device\n");
		}
		break;
	case GET_MIN_TEMP:
		if (get_data('r', val)) {
			printf("Alert Minimum Temperature Threshold: %f\n", val);
		}
		else {
			printf("ERROR: Ack not received from device\n");
		}
		break;
	case GET_MAX_TEMP:
		if (get_data('q', val)) {
			printf("Alert Maximum Temperature Threshold: %f\n", val);
		}
		else {
			printf("ERROR: Ack not received from device\n");
		}
		break;
	case GET_PERIOD:
		if (get_data('p', val)) {
			printf("Temperature Sample Period (seconds): %f\n", val);
		}
		else {
			printf("ERROR: Ack not received from device\n");
		}
		break;
	case GET_TEMP_NUM_SAMPLES:
		if (get_data('o', val)) {
			printf("Size of Temperature Sample Buffer: %d\n",(int)val);
		}
		else {
			printf("ERROR: Ack not received from device\n");
		}
		break;
	case GET_MOTION_NUM_SAMPLES:
		if (get_data('n', val)) {
			printf("Size of Motion Sample Buffer: %d\n", (int)val);
		}
		else {
			printf("ERROR: Ack not received from device\n");
		}
		break;
	case GET_MOTION_THRESH:
		if (get_data('j', val)) {
			printf("Motion Threshold X: %f\n", val);
			printf("Motion Threshold Y: %f\n",get_sample());
			printf("Motion Threshold Z: %f\n",get_sample());
		}
		else {
			printf("ERROR: Ack not received from device\n");
		}
		break;
	case TIMEOUT:
		return true;
	case EXIT:
		exit(0);
		break;
	case ACK:
		// FIXME - For protocol debug
		dev_send('k');
		break;
	default:
		return false;
	}
}